# 导出折射率IOR、高光specular、光泽sheen

import bpy

extension_is_required = False

class MaterialPropertiesProperties(bpy.types.PropertyGroup):
    enabled: bpy.props.BoolProperty(
        name = 'IOR/specular/sheen',
        description = '导出折射率IOR、高光specular、光泽sheen',
        default = True
    )

def register():
    bpy.utils.register_class(MaterialPropertiesProperties)
    bpy.types.Scene.MaterialPropertiesProperties = bpy.props.PointerProperty(type = MaterialPropertiesProperties)


def unregister():
    bpy.utils.unregister_class(MaterialPropertiesProperties)
    del bpy.types.Scene.MaterialPropertiesProperties


def pannel_draw(panel, context):
    layout = panel.layout
    layout.prop(bpy.context.scene.MaterialPropertiesProperties, 'enabled')


from io_scene_gltf2.blender.exp import gltf2_blender_get, gltf2_blender_gather_texture_info, gltf2_blender_search_node_tree

def __has_image_node_from_socket(socket):
    result = gltf2_blender_search_node_tree.from_socket(
        socket,
        gltf2_blender_search_node_tree.FilterByType(bpy.types.ShaderNodeTexImage))
    if not result:
        return False
    return True

# https://github.com/KhronosGroup/glTF-Blender-IO/pull/1158/files
def gather_ior_and_specular_extensions(blender_material, export_settings):
    lerp = lambda a, b, v: (1-v)*a + v*b
    luminance = lambda c: 0.3 * c[0] + 0.6 * c[1] + 0.1 * c[2]

    specular_ext_enabled = False
    ior_ext_enabled = False

    specular_extension = {}
    ior_extension = {}

    specular_socket = gltf2_blender_get.get_socket(blender_material, 'Specular')
    specular_tint_socket = gltf2_blender_get.get_socket(blender_material, 'Specular Tint')
    base_color_socket = gltf2_blender_get.get_socket(blender_material, 'Base Color')
    transmission_socket = gltf2_blender_get.get_socket(blender_material, 'Transmission')
    ior_socket = gltf2_blender_get.get_socket(blender_material, 'IOR')

    specular_not_linked = isinstance(specular_socket, bpy.types.NodeSocket) and not specular_socket.is_linked
    specular_tint_not_linked = isinstance(specular_tint_socket, bpy.types.NodeSocket) and not specular_tint_socket.is_linked
    base_color_not_linked = isinstance(base_color_socket, bpy.types.NodeSocket) and not base_color_socket.is_linked
    transmission_not_linked = isinstance(transmission_socket, bpy.types.NodeSocket) and not transmission_socket.is_linked
    ior_not_linked = isinstance(ior_socket, bpy.types.NodeSocket) and not ior_socket.is_linked

    specular = specular_socket.default_value if specular_not_linked else None
    specular_tint = specular_tint_socket.default_value if specular_tint_not_linked else None
    transmission = transmission_socket.default_value if transmission_not_linked else None
    ior = ior_socket.default_value if ior_not_linked else 1.5   # textures not supported

    no_texture = (specular_not_linked and specular_tint_not_linked and
        (specular_tint == 0.0 or (specular_tint != 0.0 and base_color_not_linked)))

    if ior != 1.5:
        ior_ext_enabled = True
        ior_extension['ior'] = ior

    if no_texture:
        if specular != 0.5 or specular_tint != 0.0:
            specular_ext_enabled = True
            base_color = base_color_socket.default_value[0:3] if base_color_not_linked else [0, 0, 0]
            normalized_base_color = [bc / luminance(base_color) if luminance(base_color) > 0 else 0 for bc in base_color]
            specular_color = [lerp(1, bc, specular_tint) for bc in normalized_base_color]

            # The IOR dictates the maximal reflection strength, therefore we need to clamp
            # reflection strenth of non-transmissive (aka plastic) fraction (if any)
            plastic = [1/((ior - 1) / (ior + 1))**2 * 0.08 * specular * sc for sc in specular_color]
            glass = specular_color
            specular_extension['specularColorFactor'] = [lerp(plastic[c], glass[c], transmission) for c in range(0,3)]
    else:
        sockets = (specular_socket, specular_tint_socket, base_color_socket, transmission_socket, ior_socket)
        primary_socket = specular_socket
        if specular_not_linked:
            primary_socket = specular_tint_socket
        if specular_tint_not_linked:
            primary_socket = base_color_socket
        if base_color_not_linked:
            primary_socket = transmission_socket
        info = gltf2_blender_gather_texture_info.gather_texture_info(primary_socket, sockets, export_settings)
        if info is None:
            return None, None

        specular_ext_enabled = True
        specular_extension['specularColorTexture'] = info

        # If specular>0.5, specular color may be >1. As we cannot store values >1 in a byte texture,
        # we use the specular color factor to rescale the value range.
        specular_extension['specularColorFactor'] = [2, 2, 2]

    ior_extension = ior_extension
    specular_extension = specular_extension

    return ior_extension, specular_extension

# https://github.com/KhronosGroup/glTF-Blender-IO/pull/1344/files
def gather_sheen_extension(blender_material, export_settings):
    sheen_enabled = False
    sheen_extension = {}

    sheen_socket = gltf2_blender_get.get_socket(blender_material, 'Sheen')
    sheen_tint_socket = gltf2_blender_get.get_socket(blender_material, 'Sheen Tint')
    base_color_socket = gltf2_blender_get.get_socket(blender_material, 'Base Color')
    if (isinstance(sheen_socket, bpy.types.NodeSocket)
            and isinstance(sheen_tint_socket, bpy.types.NodeSocket)
            and isinstance(base_color_socket, bpy.types.NodeSocket)):
        sheen = gltf2_blender_get.get_factor_from_socket(sheen_socket, 'VALUE') or 0
        tint = gltf2_blender_get.get_factor_from_socket(sheen_tint_socket, 'VALUE') or 0
        base_color = gltf2_blender_get.get_factor_from_socket(base_color_socket, 'RGB') or [1, 1, 1]
        sheen_extension['sheenColorFactor'] = [
            sheen * ((1 - tint) + (tint * base_color[0])),
            sheen * ((1 - tint) + (tint * base_color[1])),
            sheen * ((1 - tint) + (tint * base_color[2]))
        ]

    if not any(sheen_extension['sheenColorFactor']):
        return None

    roughness_socket = gltf2_blender_get.get_socket(blender_material, 'Roughness')
    if isinstance(roughness_socket, bpy.types.NodeSocket) and not roughness_socket.is_linked:
        sheen_extension['sheenRoughnessFactor'] = gltf2_blender_get.get_factor_from_socket(roughness_socket, 'VALUE')

    return sheen_extension



def gather_emissiveIntensity_extension(blender_material, export_settings):
    emissiveIntensitySocket = gltf2_blender_get.get_socket(blender_material, 'Emission Strength')
    emissiveIntensityExtension = {}
    emissiveIntensityExtension['emissiveIntensity'] = emissiveIntensitySocket.default_value
    return emissiveIntensityExtension



def gather_material_hook(user_extension, gltf2_material, blender_material, export_settings):
    if not bpy.context.scene.MaterialPropertiesProperties.enabled: return

    ior_extension, specular_extension = gather_ior_and_specular_extensions(blender_material, export_settings)
    sheen_extension = gather_sheen_extension(blender_material, export_settings)
    emissiveIntensityExtension = gather_emissiveIntensity_extension(blender_material, export_settings)

    if ior_extension:
        gltf2_material.extensions["KHR_materials_ior"] = user_extension.Extension(
            name = "KHR_materials_ior",
            extension = ior_extension,
            required = extension_is_required
        )
    if specular_extension:
        gltf2_material.extensions["KHR_materials_specular"] = user_extension.Extension(
            name = "KHR_materials_specular",
            extension = specular_extension,
            required = extension_is_required
        )
    if sheen_extension:
        gltf2_material.extensions["KHR_materials_sheen"] = user_extension.Extension(
            name = "KHR_materials_sheen",
            extension = sheen_extension,
            required = extension_is_required
        )
    if emissiveIntensityExtension:
        gltf2_material.extensions["DS_emissive_intensity"] = user_extension.Extension(
            name = "DS_emissive_intensity",
            extension = emissiveIntensityExtension,
            required = extension_is_required
        )