# partly support Mix Shader node in material's node tree
import bpy
import base64
import zlib
import struct

extension_is_required = False
glTF_extension_name = "DS_shader_mix"

def img_to_png(blender_image):
    '''
    code from https://blender.stackexchange.com/questions/62072/does-blender-have-a-method-to-a-get-png-formatted-bytearray-for-an-image-via-pyt
    '''
    width = blender_image.size[0]
    height = blender_image.size[1]
    buf = bytearray([int(p * 255) for p in blender_image.pixels])

    # reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4
    raw_data = b''.join(b'\x00' + buf[span:span + width_byte_4]
                        for span in range((height - 1) * width_byte_4, -1, - width_byte_4))

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (struct.pack("!I", len(data)) +
                chunk_head +
                struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head)))

    png_bytes = b''.join([
        b'\x89PNG\r\n\x1a\n',
        png_pack(b'IHDR', struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
        png_pack(b'IDAT', zlib.compress(raw_data, 9)),
        png_pack(b'IEND', b'')])

    return 'data:image/png;base64,' + base64.b64encode(png_bytes).decode()

def getTextureInfos(imageNode):
    '''
    get the properties including uscale / vscale / uoffset / voffset etc.
    '''
    textureVectorLinks = imageNode.inputs[0].links
    if len(textureVectorLinks) > 0 and textureVectorLinks[0].from_node.type == 'MAPPING':
        mappingNode = textureVectorLinks[0].from_node
        transformation = mappingNode.inputs[1].default_value
        scaling = mappingNode.inputs[3].default_value
        # CAUTION: y should be inverted
        transformation = [transformation[0], -transformation[1]]
        scaling = [scaling[0], -scaling[1]]
        return {
            'offset': transformation,
            'scaling': scaling
        }
    else:
        return {
            'offset': [0, 0],
            'scaling': [1, 1]
        }

def getPBRInfos(BSDFNode):
    '''
    returns an object that contains infos like albedoTexture / metallic
    '''
    socketName2channelName = {
        'R': 'r',
        'G': 'g',
        'B': 'b'
    }

    infos = {}

    # base color
    baseColorInfo = {}
    baseColorInput = BSDFNode.inputs[0]
    if len(baseColorInput.links) == 0:
        # color is used
        color = []
        for i in range(0, len(baseColorInput.default_value)):
            color.append(baseColorInput.default_value[i])
        baseColorInfo['type'] = 'rgba'
        baseColorInfo['val'] = color
    else:
        # texture is used
        imageNode = baseColorInput.links[0].from_node
        baseColorInfo['type'] = 'texture'
        baseColorInfo['val'] = img_to_png(imageNode.image)
        baseColorInfo['uvInfo'] = getTextureInfos(imageNode)
    infos['baseColor'] = baseColorInfo

    # metallic
    metallicInfo = {}
    metallicInput = BSDFNode.inputs[4]
    if len(metallicInput.links) == 0:
        metallicInfo['type'] = 'number'
        metallicInfo['val'] = metallicInput.default_value
    else:
        separateOutputSocketName = metallicInput.links[0].from_socket.name
        metallicUsedChannelName = socketName2channelName[separateOutputSocketName]
        separateNode = metallicInput.links[0].from_node
        imageNode = separateNode.inputs[0].links[0].from_node
        metallicInfo['type'] = 'texture'
        metallicInfo['val'] = img_to_png(imageNode.image)
        metallicInfo['uvInfo'] = getTextureInfos(imageNode)
        metallicInfo['channel'] = metallicUsedChannelName
    infos['metallic'] = metallicInfo

    # roughness
    roughnessInfo = {}
    roughnessInput = BSDFNode.inputs[7]
    if len(roughnessInput.links) == 0:
        roughnessInfo['type'] = 'number'
        roughnessInfo['val'] = roughnessInput.default_value
    else:
        separateOutputSocketName = roughnessInput.links[0].from_socket.name
        metallicUsedChannelName = socketName2channelName[separateOutputSocketName]
        separateNode = roughnessInput.links[0].from_node
        imageNode = separateNode.inputs[0].links[0].from_node
        roughnessInfo['type'] = 'texture'
        roughnessInfo['val'] = img_to_png(imageNode.image)
        roughnessInfo['uvInfo'] = getTextureInfos(imageNode)
        roughnessInfo['channel'] = metallicUsedChannelName
    infos['roughness'] = roughnessInfo

    # todo: emission (because it is hard to archive emissiveColor with glowlayer using nodeMaterial, the emission is not supported now)
    # emissiveInfo = {}
    # emissiveInput = BSDFNode.inputs[17]
    # if len(emissiveInput.links) == 0:
    #     # color is used
    #     color = []
    #     for i in range(0, len(emissiveInput.default_value)):
    #         color.append(emissiveInput.default_value[i])
    #     emissiveInfo['type'] = 'rgba'
    #     emissiveInfo['val'] = color
    # else:
    #     # texture is used
    #     imageNode = emissiveInput.links[0].from_node
    #     emissiveInfo['type'] = 'texture'
    #     emissiveInfo['val'] = img_to_png(imageNode.image)
    #     emissiveInfo['uvInfo'] = getTextureInfos(imageNode)
    # infos['emissive'] = emissiveInfo

    # normal color
    normalInfo = {}
    normalInput = BSDFNode.inputs[20]
    if len(normalInput.links) == 0:
        normalInfo['type'] = 'rgba'
        normalInfo['val'] = [0.5, 0.5, 1, 1]
    else:
        node = normalInput.links[0].from_node
        if node.type == 'NORMAL_MAP':
            normalMapNodeInput = node.inputs[1]
            imageNode = normalMapNodeInput.links[0].from_node
        elif node.type == 'TEX_IMAGE':
            imageNode = node
        assert imageNode, 'Can not find the image node for the normal!'
        normalInfo['type'] = 'texture'
        normalInfo['val'] = img_to_png(imageNode.image)
        normalInfo['uvInfo'] = getTextureInfos(imageNode)
    infos['normal'] = normalInfo

    return infos

def getFactorInfos(separateNode, channelSocket):
    '''
    returns an object that contains infos with the vertex color's name and channel
    CAUTION: I will assume that a vertex color node is linked to the separate node
    '''
    socketName2channelName = {
        'R': 'red',
        'G': 'green',
        'B': 'blue'
    }

    infos = {'channel': socketName2channelName[channelSocket.name]}
    separateNodeInputLinks = separateNode.inputs[0].links
    assert len(separateNodeInputLinks) > 0, 'input of the separate node is not in use'
    vertexColorNode = separateNodeInputLinks[0].from_node
    infos['attributeDataName'] = vertexColorNode.layer_name

    return infos

def getBSDFNodesWithFactor(mixNode):
    if not mixNode:
        return
    returnObj = {}
    inputFactorLinks = mixNode.inputs[0].links
    inputLeftLinks = mixNode.inputs[1].links
    inputRightLinks = mixNode.inputs[2].links
    mixNodeAllInputInUseFlag = len(inputFactorLinks) > 0 and len(inputLeftLinks) > 0 and len(inputRightLinks) > 0
    assert mixNodeAllInputInUseFlag, 'some inputs of ' + mixNode.name + ' is not in use'

    returnObj['factor'] = getFactorInfos(inputFactorLinks[0].from_node, inputFactorLinks[0].from_socket)
    leftNode = inputLeftLinks[0].from_node
    if leftNode.type == 'MIX_SHADER':
        returnObj['left'] = getBSDFNodesWithFactor(leftNode)
    else:
        returnObj['left'] = getPBRInfos(leftNode)
    rightNode = inputRightLinks[0].from_node
    if rightNode.type == 'MIX_SHADER':
        returnObj['right'] = getBSDFNodesWithFactor(rightNode)
    else:
        returnObj['right'] = getPBRInfos(rightNode)
    return returnObj



# hooks
def gather_material_hook(user_extension, gltf2_material, blender_material, export_settings):
    lastMixNode = None
    for node in blender_material.node_tree.nodes:
        if node.type != 'MIX_SHADER':
            continue
        outputLinks = node.outputs[0].links
        if len(outputLinks) > 0:
            nextNodeName = outputLinks[0].to_node.name
            if nextNodeName == 'Material Output':
                lastMixNode = node
                break

    extensionObj = getBSDFNodesWithFactor(lastMixNode)
    if extensionObj:
        blender_material['isMixMat'] = True
        gltf2_material.extensions[glTF_extension_name] = user_extension.Extension(
            name = glTF_extension_name,
            extension = extensionObj,
            required = extension_is_required
        )
    else:
        blender_material['isMixMat'] = False

def gather_mesh_hook(user_extension, gltf2_mesh, blender_mesh, blender_object, vertex_groups, modifiers, skip_filter, material_names, export_settings):
    vertexColorNames = []
    vertexColors = blender_mesh.vertex_colors
    for i in range(0, len(vertexColors)):
        vertexColorName = vertexColors[i].name
        vertexColorNames.append(vertexColorName)

    hasMixMatFlag = False
    for mat in blender_mesh.materials:
        if mat['isMixMat']:
            hasMixMatFlag = True
            break

    if len(vertexColors) > 0 and hasMixMatFlag:
        gltf2_mesh.extensions[glTF_extension_name] = user_extension.Extension(
            name = glTF_extension_name,
            extension = {
                'vertexColorNames': vertexColorNames
            },
            required = extension_is_required
        )